"""
An adaptation of Andrej Karpathy's nanoGPT implementation in PyTorch.
Original source: https://github.com/karpathy/nanoGPT

Original License:
MIT License

Copyright (c) 2022 Andrej Karpathy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Original comments:
Full definition of a GPT Language Model, all of it in this single file.
References:
1) the official GPT-2 TensorFlow implementation released by OpenAI:
https://github.com/openai/gpt-2/blob/master/src/model.py
2) huggingface/transformers PyTorch implementation:
https://github.com/huggingface/transformers/blob/main/src/transformers/models/gpt2/modeling_gpt2.py
"""

import math
from dataclasses import dataclass

import torch
import torch.nn as nn
from torch.nn import functional as F


# @torch.jit.script # good to enable when not using torch.compile, disable when using (our default)
def new_gelu(x):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """
    return (
        0.5
        * x
        * (
            1.0
            + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0)))
        )
    )


class CausalSelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd)
        # output projection
        self.c_proj = nn.Linear(config.n_embd, config.n_embd)
        # regularization
        self.attn_dropout = nn.Dropout(config.dropout)
        self.resid_dropout = nn.Dropout(config.dropout)
        # causal mask to ensure that attention is only applied to the left in the input sequence
        self.register_buffer(
            "bias",
            torch.tril(torch.ones(config.block_size, config.block_size)).view(
                1, 1, config.block_size, config.block_size
            ),
        )
        self.n_head = config.n_head
        self.n_embd = config.n_embd

    def forward(self, x):
        # batch size, sequence length, embedding dimensionality (n_embd)
        B, T, C = x.size()

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float("-inf"))
        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y


class MLP(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.c_fc = nn.Linear(config.n_embd, 4 * config.n_embd)
        self.c_proj = nn.Linear(4 * config.n_embd, config.n_embd)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, x):
        x = self.c_fc(x)
        x = new_gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)
        return x


class Block(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.n_embd)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = nn.LayerNorm(config.n_embd)
        self.mlp = MLP(config)

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


@dataclass
class GPTConfig:
    block_size: int = 1024
    input_dim: int = 256
    output_dim: int = 256
    n_layer: int = 12
    n_head: int = 12
    n_embd: int = 768
    dropout: float = 0.1


class GPT(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.input_dim is not None
        assert config.output_dim is not None
        assert config.block_size is not None
        self.config = config

        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Linear(config.input_dim, config.n_embd),
                wpe=nn.Embedding(config.block_size, config.n_embd),
                drop=nn.Dropout(config.dropout),
                h=nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
                ln_f=nn.LayerNorm(config.n_embd),
            )
        )
        self.lm_head = nn.Linear(config.n_embd, config.output_dim, bias=False)
        # init all weights, and apply a special scaled init to the residual projections, per GPT-2 paper
        self.apply(self._init_weights)
        for pn, p in self.named_parameters():
            if pn.endswith("c_proj.weight"):
                torch.nn.init.normal_(
                    p, mean=0.0, std=0.02 / math.sqrt(2 * config.n_layer)
                )

        # report number of parameters
        n_params = sum(p.numel() for p in self.parameters())

    def forward(self, input, targets=None):
        device = input.device
        b, t, d = input.size()
        assert (
            t <= self.config.block_size
        ), f"Cannot forward sequence of length {t}, block size is only {self.config.block_size}"
        pos = torch.arange(0, t, dtype=torch.long, device=device).unsqueeze(
            0
        )  # shape (1, t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(
            input
        )  # token embeddings of shape (b, t, n_embd)
        pos_emb = self.transformer.wpe(
            pos
        )  # position embeddings of shape (1, t, n_embd)
        x = self.transformer.drop(tok_emb + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)
        logits = self.lm_head(x)
        return logits

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
        elif isinstance(module, nn.LayerNorm):
            torch.nn.init.zeros_(module.bias)
            torch.nn.init.ones_(module.weight)

    def crop_block_size(self, block_size):
        assert block_size <= self.config.block_size
        self.config.block_size = block_size
        self.transformer.wpe.weight = nn.Parameter(
            self.transformer.wpe.weight[:block_size]
        )
        for block in self.transformer.h:
            block.attn.bias = block.attn.bias[:, :, :block_size, :block_size]

    def configure_optimizers(self, weight_decay, learning_rate, betas):
        """
        This long function is unfortunately doing something very simple and is being very defensive:
        We are separating out all parameters of the model into two buckets: those that will experience
        weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
        We are then returning the PyTorch optimizer object.
        """

        # separate out all parameters to those that will and won't experience regularizing weight decay
        decay = set()
        no_decay = set()
        whitelist_weight_modules = (torch.nn.Linear,)
        blacklist_weight_modules = (torch.nn.LayerNorm, torch.nn.Embedding)
        for mn, m in self.named_modules():
            for pn, p in m.named_parameters():
                fpn = "%s.%s" % (mn, pn) if mn else pn  # full param name
                if pn.endswith("bias"):
                    # all biases will not be decayed
                    no_decay.add(fpn)
                elif pn.endswith("weight") and isinstance(m, whitelist_weight_modules):
                    # weights of whitelist modules will be weight decayed
                    decay.add(fpn)
                elif pn.endswith("weight") and isinstance(m, blacklist_weight_modules):
                    # weights of blacklist modules will NOT be weight decayed
                    no_decay.add(fpn)

        # validate that we considered every parameter
        param_dict = {pn: p for pn, p in self.named_parameters()}
        inter_params = decay & no_decay
        union_params = decay | no_decay
        assert len(inter_params) == 0, (
            "parameters %s made it into both decay/no_decay sets!"
            % (str(inter_params),)
        )
        assert len(param_dict.keys() - union_params) == 0, (
            "parameters %s were not separated into either decay/no_decay set!"
            % (str(param_dict.keys() - union_params),)
        )

        # create the pytorch optimizer object
        optim_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(decay))],
                "weight_decay": weight_decay,
            },
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay))],
                "weight_decay": 0.0,
            },
        ]
        optimizer = torch.optim.AdamW(optim_groups, lr=learning_rate, betas=betas)
        return optimizer



import torch
import numpy as np
import torch.nn as nn
import os.path as osp


def weights_init_encoder(m):
    if isinstance(m, nn.Linear):
        nn.init.orthogonal_(m.weight.data)
        m.bias.data.fill_(0.0)
    elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
        assert m.weight.size(2) == m.weight.size(3)
        m.weight.data.fill_(0.0)
        m.bias.data.fill_(0.0)
        mid = m.weight.size(2) // 2
        gain = nn.init.calculate_gain("relu")
        nn.init.orthogonal_(m.weight.data[:, :, mid, mid], gain)


def get_tensor(z, device):
    if z is None:
        return None
    if z[0].dtype == np.dtype("O"):
        return None
    if len(z.shape) == 1:
        return torch.FloatTensor(z.copy()).to(device).unsqueeze(0)
        # return torch.from_numpy(z.copy()).float().to(device).unsqueeze(0)
    else:
        return torch.FloatTensor(z.copy()).to(device)
        # return torch.from_numpy(z.copy()).float().to(device)



import torch
from typing import Callable, List, Optional


class MLP(torch.nn.Sequential):
    """This block implements the multi-layer perceptron (MLP) module.
    Adapted for backward compatibility from the torchvision library:
    https://pytorch.org/vision/0.14/generated/torchvision.ops.MLP.html

    LICENSE:

    From PyTorch:

    Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
    Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
    Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
    Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
    Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
    Copyright (c) 2011-2013 NYU                      (Clement Farabet)
    Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
    Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
    Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

    From Caffe2:

    Copyright (c) 2016-present, Facebook Inc. All rights reserved.

    All contributions by Facebook:
    Copyright (c) 2016 Facebook Inc.

    All contributions by Google:
    Copyright (c) 2015 Google Inc.
    All rights reserved.

    All contributions by Yangqing Jia:
    Copyright (c) 2015 Yangqing Jia
    All rights reserved.

    All contributions by Kakao Brain:
    Copyright 2019-2020 Kakao Brain

    All contributions by Cruise LLC:
    Copyright (c) 2022 Cruise LLC.
    All rights reserved.

    All contributions from Caffe:
    Copyright(c) 2013, 2014, 2015, the respective contributors
    All rights reserved.

    All other contributions:
    Copyright(c) 2015, 2016 the respective contributors
    All rights reserved.

    Caffe2 uses a copyright model similar to Caffe: each contributor holds
    copyright over their contributions to Caffe2. The project versioning records
    all such contribution and copyright details. If a contributor wants to further
    mark their specific copyright on a particular contribution, they should
    indicate their copyright solely in the commit message of the change when it is
    committed.

    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
    and IDIAP Research Institute nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.


    Args:
        in_channels (int): Number of channels of the input
        hidden_channels (List[int]): List of the hidden channel dimensions
        norm_layer (Callable[..., torch.nn.Module], optional): Norm layer that will be stacked on top of the linear layer. If ``None`` this layer won't be used. Default: ``None``
        activation_layer (Callable[..., torch.nn.Module], optional): Activation function which will be stacked on top of the normalization layer (if not None), otherwise on top of the linear layer. If ``None`` this layer won't be used. Default: ``torch.nn.ReLU``
        inplace (bool, optional): Parameter for the activation layer, which can optionally do the operation in-place.
            Default is ``None``, which uses the respective default values of the ``activation_layer`` and Dropout layer.
        bias (bool): Whether to use bias in the linear layer. Default ``True``
        dropout (float): The probability for the dropout layer. Default: 0.0
    """

    def __init__(
        self,
        in_channels: int,
        hidden_channels: List[int],
        activation_layer: Optional[Callable[..., torch.nn.Module]] = torch.nn.ReLU,
        inplace: Optional[bool] = None,
        bias: bool = True,
        dropout: float = 0.0,
    ):
        params = {} if inplace is None else {"inplace": inplace}

        layers = []
        in_dim = in_channels
        for hidden_dim in hidden_channels[:-1]:
            layers.append(torch.nn.Linear(in_dim, hidden_dim, bias=bias))
            layers.append(activation_layer(**params))
            layers.append(torch.nn.Dropout(dropout, **params))
            in_dim = hidden_dim

        layers.append(torch.nn.Linear(in_dim, hidden_channels[-1], bias=bias))
        layers.append(torch.nn.Dropout(dropout, **params))

        super().__init__(*layers)



import collections

import einops
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import torch


def generate_plots(ground_truths, predictions, sampled_images, to_plot=8, traj_index=0):
    """
    Generates plots comparing ground truth vs predictions and displays sampled images.

    Parameters:
    - ground_truths: A numpy array of shape (T, 7) containing ground truth actions.
    - predictions: A numpy array of shape (T, 7) with model's predictions.
    - sampled_images: A numpy array containing M sampled images.
    """

    # Initialize the figure
    fig = plt.figure(figsize=(15, 15))
    outer_grid = gridspec.GridSpec(8, 1, hspace=0.25, wspace=0.0)

    # Top grid for images
    M = sampled_images.shape[0]
    chosen_indices = np.linspace(0, M - 1, to_plot, dtype=int)
    chosen_images = sampled_images[chosen_indices]
    top_grid = gridspec.GridSpecFromSubplotSpec(
        1, len(chosen_images), subplot_spec=outer_grid[0, :], wspace=0.0
    )

    for i, img in enumerate(chosen_images):
        ax = plt.Subplot(fig, top_grid[i])
        ax.imshow(img)
        ax.axis("off")
        if i > 0:  # To further remove any potential whitespace
            ax.set_yticklabels([])
            ax.set_xticklabels([])
            ax.set_xticks([])
            ax.set_yticks([])
        fig.add_subplot(ax)

    # For each action dimension, plot ground truth vs prediction
    dim = ground_truths.shape[1]
    for i in range(dim):
        ax = plt.Subplot(fig, outer_grid[i + 1, :])

        gt_values = ground_truths[:, i]
        pred_values = predictions[:, i]

        ax.plot(gt_values, "g", label="Ground Truth")
        ax.plot(pred_values, "r--", label="Prediction")

        ax.legend(loc="upper right")
        ax.set_title(f"Act Dim {i + 1}", fontsize=16)
        fig.add_subplot(ax)

    # save the plot
    fig.savefig(f"trajectory_{traj_index}.png")


@torch.no_grad()
def visualize_trajectory(
    model,
    test_dataset,
    device,
    buffer_size=6,
    n_visualized_trajectories=5,
    goal_conditional=False,
):
    """
    Visualizes the trajectory of the model on the test dataset.
    M images are sampled from the test dataset and displayed.
    T = end - start is the length of the trajectory.
    start and end (index) indicate the window of the trajectory to visualize.
    """
    print(goal_conditional)
    action_preds = []
    ground_truth = []
    images = []
    image_buffer = collections.deque(maxlen=buffer_size)
    test_dataset.set_include_trajectory_end(True)

    i = 0
    done_visualizing = 0
    while (done_visualizing < n_visualized_trajectories) and (i < len(test_dataset)):
        if goal_conditional:
            (input_images, terminate), goals, *_, gt_actions = test_dataset[i]
        else:
            (input_images, terminate), *_, gt_actions = test_dataset[i]
        input_images = input_images.float() / 255.0
        image_buffer.append(input_images[-1])
        img = input_images[-1]
        images.append(einops.rearrange(img, "c h w -> h w c").cpu().detach().numpy())
        ground_truth.append(gt_actions[-1])
        if goal_conditional:
            model_input = (
                torch.stack(tuple(image_buffer), dim=0).unsqueeze(0).to(device),
                torch.tensor(goals).unsqueeze(0).to(device),
                torch.tensor(gt_actions).unsqueeze(0).to(device),
            )
        else:
            model_input = (
                torch.stack(tuple(image_buffer), dim=0).unsqueeze(0).to(device),
                torch.tensor(gt_actions).unsqueeze(0).to(device),
            )

        out, _ = model.step(model_input)
        action_preds.append(out.squeeze().cpu().detach().numpy())

        if terminate:
            action_preds = np.array(action_preds)
            ground_truth = np.array(ground_truth)
            images = np.array(images)

            print(action_preds.shape, ground_truth.shape, images.shape)

            generate_plots(
                ground_truth, action_preds, images, traj_index=done_visualizing
            )
            done_visualizing += 1
            # Reset everything.
            action_preds = []
            ground_truth = []
            images = []
            image_buffer = collections.deque(maxlen=buffer_size)

        i += 1




import numpy as np
import pandas as pd
import altair as alt

# Register custom font for export
alt.themes.register('custom_theme', lambda: {
    'config': {
        'title': {'font': 'Produkt'},
        'axis': {'labelFont': 'Produkt', 'titleFont': 'Produkt'},
        'legend': {'labelFont': 'Produkt', 'titleFont': 'Produkt'},
        'mark': {'font': 'Produkt'},
        'text': {'font': 'Produkt'},
    }
})
alt.themes.enable('custom_theme')

# Hardcoded data from experiments
MODEL_DATA = {
    "CAP + Oracle": {
        "total": 5000,
        "outcomes": {
            "Success": 3994,
            "Did not lift enough": 89,
            "Object touched but not grasped": 412,
            "Picked wrong object": 59,
            "Empty Grasp": 446,
            "Did not grasp": 0,
        }
    },
    "CAP + Gemini": {
        "total": 5000,
        "outcomes": {
            "Success": 3441,
            "Did not lift enough": 88,
            "Object touched but not grasped": 389,
            "Picked wrong object": 448,
            "Empty Grasp": 634,
            "Did not grasp": 0,
        }
    },
    "CAP + Molmo": {
        "total": 5000,
        "outcomes": {
            "Success": 3008,
            "Did not lift enough": 94,
            "Object touched but not grasped": 467,
            "Picked wrong object": 877,
            "Empty Grasp": 554,
            "Did not grasp": 0,
        }
    },
    "CAP + Moondream": {
        "total": 5000,
        "outcomes": {
            "Success": 3033,
            "Did not lift enough": 83,
            "Object touched but not grasped": 342,
            "Picked wrong object": 746,
            "Empty Grasp": 796,
            "Did not grasp": 0,
        }
    }
}


def plot_failure_modes():
    model_names = list(MODEL_DATA.keys())
    
    # Print results to terminal
    print("\n" + "="*80)
    print("WRONG OBJECT SELECTION RATE BY VLM (4 Distractors)")
    print("="*80)
    for model in model_names:
        data = MODEL_DATA[model]
        total = data["total"]
        print(f"\n{model} (n={total}):")
        print("-" * 60)
        for outcome, count in data["outcomes"].items():
            percentage = (count / total * 100) if total > 0 else 0
            print(f"  {outcome:40s}: {count:3d} ({percentage:5.1f}%)")
    print("\n" + "="*80 + "\n")
    
    # Only show "Picked wrong object" percentage
    chart_data = []
    for model in model_names:
        data = MODEL_DATA[model]
        total = data["total"] or 1
        count = data["outcomes"].get("Picked wrong object", 0)
        proportion = count / total
        percentage = proportion * 100
        
        # Calculate standard error for binomial proportion
        se = np.sqrt(proportion * (1 - proportion) / total) * 100
        
        chart_data.append({
            'Model': model,
            'Percentage': percentage,
            'SE': se,
        })
    
    df = pd.DataFrame(chart_data)
    
    # Sort by percentage (lowest to highest)
    df = df.sort_values('Percentage')
    
    # Create ordered list for x-axis
    model_order = df['Model'].tolist()
    
    # Assign colors from purple family (lightest for best, darkest for worst)
    purple_shades = ["#D6BAE2", "#9B66BB", "#7B4FA2", "#4B136D"]  # Lightest to darkest
    color_mapping = {model: purple_shades[i] for i, model in enumerate(model_order)}
    df['Color'] = df['Model'].map(color_mapping)
    
    # Create simple bar chart with gradient colors
    bars = alt.Chart(df).mark_bar(
        stroke='white',
        strokeWidth=1,
        cornerRadiusTopLeft=8,
        cornerRadiusTopRight=8
    ).encode(
        x=alt.X('Model:N', title=None, sort=model_order, axis=alt.Axis(labelFontSize=18, labelAngle=0)),
        y=alt.Y('Percentage:Q', title=None, axis=alt.Axis(labelFontSize=18)),
        color=alt.Color('Model:N', scale=alt.Scale(domain=model_order, range=[color_mapping[m] for m in model_order]), legend=None),
        tooltip=['Model', alt.Tooltip('Percentage:Q', format='.1f')]
    )
    
    # Add error bars
    error_bars = alt.Chart(df).mark_errorbar(
        color='black',
        thickness=2,
        ticks=alt.MarkConfig(width=10)
    ).encode(
        x=alt.X('Model:N', sort=model_order),
        y=alt.Y('Percentage:Q'),
        yError=alt.YError('SE:Q')
    )
    
    # Add percentage labels on top of bars
    text = alt.Chart(df).mark_text(
        align='center',
        baseline='bottom',
        dy=-15,
        fontSize=16,
        fontWeight='bold'
    ).encode(
        x=alt.X('Model:N', sort=model_order),
        y=alt.Y('Percentage:Q'),
        text=alt.Text('Percentage:Q', format='.0f')
    ).transform_calculate(
        label='datum.Percentage + "%"'
    )
    
    chart = (bars + error_bars + text).properties(
        width=600,
        height=400,
        title={
            'text': '       Wrong Object Selection Rate by VLM (4 Distractors)',
            'fontSize': 22,
            'anchor': 'start',
            'dx': 60,
            'dy': -20
        },
        padding={"left": 5, "right": 5, "top": 20, "bottom": 0}
    ).configure_view(
        strokeWidth=0
    )
    
    # Save chart
    chart.save("failure_modes_by_model.html")
    chart.save("failure_modes_by_model.png", scale_factor=3)
    chart.save("failure_modes_by_model.pdf")
    print("Saved: failure_modes_by_model.html, .png, .pdf")


if __name__ == "__main__":
    plot_failure_modes()



from typing import Optional, Sequence, Tuple

import einops
import torch
import torch.nn as nn

from baselines.rum.loss_fns.abstract_loss_fn import AbstractLossFn
from baselines.rum.models.bet import (
    GPT,
    BehaviorTransformer,
    TokenizedBehaviorTransformer,
)
from baselines.rum.models.bet.vqvae.vqvae import VqVae


class RVQLossFn(AbstractLossFn):
    def __init__(
        self,
        tokenized_bet: bool,
        action_dim: int,
        xyz_only: bool,
        mask_last_min: int = 0,
        mask_last_max: int = 0,
        gpt_input_dim: int = 0,
        learned_mask: bool = True,
        model: Optional[torch.nn.Module] = None,
        predict_with_offsets: bool = True,
        sampling_temperature: float = 1.0,
        action_sequence_length: int = 1,
        vqvae_n_latent_dims: int = 512,
        vqvae_n_embed: int = 16,
        vqvae_groups: int = 2,
        obs_cond: bool = False,
    ):
        super().__init__(model)
        assert mask_last_max >= mask_last_min
        assert mask_last_min >= 0
        obs_dim = model.feature_dim
        self._true_action_dim = action_dim
        action_dim = 3 if xyz_only else action_dim
        self.obs_cond = obs_cond
        self._model = model
        self._rvq = VqVae(
            obs_dim=gpt_input_dim,
            input_dim_h=action_sequence_length,
            input_dim_w=action_dim,
            n_latent_dims=vqvae_n_latent_dims,
            vqvae_n_embed=vqvae_n_embed,
            vqvae_groups=vqvae_groups,
            eval=False,
            device=model.device,
            enc_loss_type="through_vqlayer",
            obs_cond=obs_cond,
        )

        if self.obs_cond:
            self._obs_mask_token = (
                nn.Parameter(torch.ones(gpt_input_dim), requires_grad=False)
                if not learned_mask or mask_last_max == 0
                else nn.Parameter(torch.randn(gpt_input_dim), requires_grad=True)
            )
            self._obs_adapter = nn.Linear(obs_dim, gpt_input_dim, bias=False)
            self._adapt_obs = (
                self._adapt_obs_linear
                if not tokenized_bet
                else self._adapt_obs_tokenized
            )
        self._mask_last = (mask_last_min, mask_last_max)
        self._action_dim = action_dim

        self.step = self._step if not tokenized_bet else self._step_tokenized

        self._seen_action_stack = []
        self._start_and_ends = None

    def _adapt_obs_tokenized(self, obs):
        return einops.rearrange(
            self._obs_adapter(einops.rearrange(obs, "... c h w -> ... h w c")),
            "... h w c -> ... c h w",
        )

    def _adapt_obs_linear(self, obs):
        return self._obs_adapter(obs)

    def _begin_epoch(self, *args, **kwargs):
        return self._rvq._begin_epoch(*args, **kwargs)

    def forward(self, data, output, *args, **kwargs):
        *_, padding, actions = data
        action_seq = actions[..., : self._action_dim].contiguous()
        if self.obs_cond:
            adapted_obs = self._adapt_obs(output)
            loss, loss_dict = self._rvq.vqvae_update(action_seq, adapted_obs)
        else:
            loss, loss_dict = self._rvq.vqvae_update(action_seq, None)
        return loss, loss_dict

    @torch.no_grad()
    def _step(self, data, output, *args, **kwargs):
        pass

    @torch.no_grad()
    def _step_tokenized(self, data, output, *args, **kwargs):
        pass



import numpy as np
import time
from egogym.policies.base_policy import BasePolicy
from egogym.utils import resize_with_pad

class PIPolicy(BasePolicy):
    
    def __init__(self, config=None, **kwargs):
        super().__init__()
        if config is None:
            config = get_config()
        
        for key, value in kwargs.items():
            config[key] = value
            
        self.config = config
        self.name = "pi"
        self.host = self.config["host"]
        self.port = self.config["port"]
        self.grasping_style = self.config["grasping_style"]
        self.buffer_length = self.config["buffer_length"]
        self.batch_size = 0

        try:
            from openpi_client import websocket_client_policy
        except ImportError:
            raise ImportError("Please install the openpi-client.")
        
        max_retries = 5
        for attempt in range(max_retries):
            try:
                self.model = websocket_client_policy.WebsocketClientPolicy(
                    host=self.host,
                    port=self.port,
                )
                print(f"Connected to PI model at {self.host}:{self.port}")
                break
            except Exception as e:
                if attempt < max_retries - 1:
                    print(f"Connection attempt {attempt + 1} failed: {e}. Retrying...")
                    time.sleep(1)
                else:
                    print(f"Failed to connect to remote model after {max_retries} attempts")
                    raise
                
    def get_action(self, obs):

        if self.batch_size == 0:
            if obs["rgb_exo"].ndim == 3:
                self.batch_size = 1
            else:
                self.batch_size = obs["rgb_exo"].shape[0]
            self.action_buffer = [None] * self.batch_size
            self.current_buffer_index = [0] * self.batch_size

        if self.batch_size == 1:
            obs = {key: np.expand_dims(value, axis=0) if isinstance(value, np.ndarray) else [value] for key, value in obs.items()}

        
        actions = []
        for i in range(self.batch_size):

            if self.action_buffer[i] is None or self.current_buffer_index[i] >= self.buffer_length:
                model_input = {
                    "observation/exterior_image_1_left": resize_with_pad(obs["rgb_exo"][i], 224, 224),
                    "observation/wrist_image_left": resize_with_pad(obs["rgb_ego"][i], 224, 224),
                    "observation/joint_position": np.array(obs["joint_positions"][i][:7]).reshape(7,),
                    "observation/gripper_position": np.array(obs["grasp"][i]).reshape(1,),
                    "prompt": f'pick up the {obs["object_name"][i]} and place it on a plate.',
                }
                self.action_buffer[i] = self.model.infer(model_input)["actions"]  
                self.current_buffer_index[i] = 0    
            current_action = self.action_buffer[i][self.current_buffer_index[i]]
            self.current_buffer_index[i] += 1

            gripper_pos = np.array([np.clip(current_action[7], 0.0, 1.0)])

            if self.grasping_style == "binary":
                gripper_pos = np.array([1.0]) if gripper_pos >= self.config["gripper_threshold"] else np.array([0.0])
            elif self.grasping_style == "semi_binary":
                gripper_pos = gripper_pos if gripper_pos <= self.config["gripper_threshold"] else np.array([1.0])
            elif not self.grasping_style == "continuous":
                raise ValueError(f"Invalid grasping style: {self.grasping_style}")
                
                
            arm_output = current_action[:7].reshape(7)
            actions.append(np.concatenate([arm_output, gripper_pos], axis=0).reshape(1, -1))

        return np.array(actions).reshape(self.batch_size, 8)
    
    def reset(self, indicies=None):
        if indicies is not None:
            for i in indicies:
                self.action_buffer[i] = None
                self.current_buffer_index[i] = 0
        else:
            self.batch_size = 0
            self.action_buffer = None
            self.current_buffer_index = None


def get_config():
    config = {
        "host": "localhost",
        "port": 8000,
        "gripper_threshold": 0.5,
        "grasping_style": "binary",
        "buffer_length": 8,
    }
    return config
